import json
import os
from typing import Dict
from typing.io import BinaryIO

# run_test.py will rewrite the variable `CONFIG_FILE_PATH` so we have to access
# the variable through the module. Don't import variable directly.
import abejacli.configuration

from abejacli.configuration.config import Config
from abejacli.exceptions import InvalidConfigException


class ConfigLoader(object):
    """
    The loader class which loads and caches user configurations.
    """

    def load(self, name: str = 'default') -> Config:
        try:
            with open(abejacli.configuration.CONFIG_FILE_PATH, 'r') as f:
                return self.load_from_file(f, name)
        except FileNotFoundError:
            # Try to read configuration from environment variables only
            return self.load_from_dict({}, name)

    def load_from_file(self, fp: BinaryIO, name: str) -> Config:
        try:
            return self.load_from_dict(json.load(fp), name)
        except json.JSONDecodeError:
            raise InvalidConfigException('malformed configuration file')

    def load_from_string(self, string: str, name: str) -> Config:
        try:
            return self.load_from_dict(json.loads(string), name)
        except json.JSONDecodeError:
            raise InvalidConfigException('malformed configuration file')

    def load_from_dict(self, plain_config: Dict[str, str], name: str) -> Config:
        named_configs = plain_config.get('configurations', {})
        config = named_configs.get(name)

        # If default config not found, use old config.
        if name == 'default' and not config:
            config = plain_config

        # Environment variable > Environment variable (obsolete) > config file
        user = os.environ.get('ABEJA_CLI_USER',
                              os.environ.get('ABEJA_PLATFORM_USER',
                                             config.get('abeja-platform-user')))
        token = os.environ.get('ABEJA_CLI_TOKEN',
                               os.environ.get('PERSONAL_ACCESS_TOKEN',
                                              config.get('personal-access-token')))
        organization = os.environ.get('ABEJA_CLI_ORGANIZATION',
                                      os.environ.get('ORGANIZATION_NAME',
                                                     config.get('organization-name')))

        # All configuration must be filled.
        variables = {
            'abeja-platform-user': user,
            'personal-access-token': token,
            'organization-name': organization
        }

        for config_key, value in variables.items():
            if value is None:
                raise InvalidConfigException(
                    'missing config value: {}'.format(config_key))

        return Config(user=user, token=token, organization=organization)
