import json
from io import StringIO
from operator import itemgetter
import os
import sys
import urllib

import click
from click.exceptions import MissingParameter
import click_config_file

from abejacli.click_custom import (DATASET_PARAM_STR, ENVIRONMENT_STR, USER_PARAM_STR, VOLUME_PARAM_STR)
from abejacli.common import json_output_formatter, progress_status, version_archive
from abejacli.common import get_organization_id
from abejacli.config import (ABEJA_PLATFORM_TOKEN, ABEJA_PLATFORM_USER_ID,
                             CONFIG, DEFAULT_EXCLUDE_FILES,
                             ERROR_EXITCODE, ORGANIZATION_ENDPOINT,
                             SUCCESS_EXITCODE)
from abejacli.exceptions import ConfigFileNotFoundError, InvalidConfigException, ResourceNotFound
from abejacli.docker.commands.run import build_volumes
from abejacli.logger import get_logger
from abejacli.session import api_get, api_post, api_patch
from abejacli.training import TrainingConfig, CONFIGFILE_NAME, read_training_config
from abejacli.training.jobs import TrainingJobDebugRun
from abejacli.training.jobs import TrainingJobLocalContainerRun

logger = get_logger()

training_config = TrainingConfig()


@click.group(help='Training operation commands')
@click.pass_context
def training(ctx):
    if not CONFIG:
        click.echo(
            "[error] there is no configuration, execute 'abeja configure'")
        ctx.abort()
        return


@training.command(name='init', help='Initialize training with training definition name')
@click.argument('name', required=True)
def initialize_training(name):
    try:
        training_config.write(name)
    except Exception as e:
        logger.error('training initialization failed: {}'.format(e))
        click.echo('training initialization failed.')
        sys.exit(ERROR_EXITCODE)
    click.echo('training initialized')


@training.command(name='create-job-definition', help='Create training job definition')
def create_job_definition():
    try:

        config_data = training_config.read(training_config.default_schema)
        params = {
            "name": config_data['name']
        }
        url = "{}/training/definitions".format(ORGANIZATION_ENDPOINT)
        json_data = json.dumps(params)
        r = api_post(url, json_data)
    except ConfigFileNotFoundError:
        click.echo('training configuration file does not exists.')
        logger.error('training configuration file does not exists.')
        sys.exit(ERROR_EXITCODE)
    except InvalidConfigException as e:
        logger.error('invalid training configuration file: {}'.format(e))
        click.echo('invalid training configuration file.')
        sys.exit(ERROR_EXITCODE)
    except Exception as e:
        logger.error('create job definition aborted:{}'.format(e))
        click.echo('create job definition aborted.')
        sys.exit(ERROR_EXITCODE)
    click.echo(json_output_formatter(r))


@training.command(name='create-version')
@click.option('-d', '--description', type=str, required=False,
              help='Description for the training job, which must be less than or equal to 256 characters.')
@click.option('-e', '--exclude', type=str, help='Specify the file/dir to exclude from create-version.',
              required=False, multiple=True)
def create_training_version(description, exclude):
    if exclude is None:
        excludes = []
    else:
        excludes = list(exclude)

    try:
        params = training_config.read(training_config.create_version_schema)
        if params.get('params'):
            params['user_parameters'] = params['params']
        if description is not None:
            params['description'] = description

        user_exclude_files = params.pop('ignores', [])
        exclude_files = set(user_exclude_files + DEFAULT_EXCLUDE_FILES + excludes)
        archive = version_archive(params['name'], exclude_files)

        url = "{}/training/definitions/{}/versions".format(
            ORGANIZATION_ENDPOINT, params['name'])
        with open(archive.name, 'rb') as archive_file:
            json_params = json.dumps(params)
            files = {
                'source_code': ('archive.tar.gz', archive_file, 'application/tar+gzip'),
                'parameters': ('params.json', StringIO(json_params), 'application/json'),
            }
            r = api_post(url, files=files)
    except ConfigFileNotFoundError:
        logger.error('training configuration file does not exists.')
        click.echo('training configuration file does not exists.')
        sys.exit(ERROR_EXITCODE)
    except FileNotFoundError:
        logger.error('failed to archive the current directory.')
        click.echo('failed to archive the current directory.')
        sys.exit(ERROR_EXITCODE)
    except InvalidConfigException as e:
        logger.error('invalid training configuration file: {}'.format(e))
        click.echo('invalid training configuration file.')
        sys.exit(ERROR_EXITCODE)
    except Exception as e:
        logger.error('create training version aborted: {}'.format(e))
        click.echo('create training version aborted.')
        sys.exit(ERROR_EXITCODE)
    finally:
        if os.path.exists(archive.name):
            archive.close()
            os.unlink(archive.name)
    click.echo(json_output_formatter(r))


@training.command(name='update-version')
@click.option('-v', '--version', type=str, required=False,
              help='Job definition version. By default, latest version is used')
@click.option('-d', '--description', type=str, required=False,
              help='Description for the training job, which must be less than or equal to 256 characters.')
def update_training_version(version, description):
    try:
        config = training_config.read(training_config.create_version_schema)
    except ConfigFileNotFoundError:
        logger.error('training configuration file does not exists.')
        click.echo('training configuration file does not exists.')
        sys.exit(ERROR_EXITCODE)
    except InvalidConfigException as e:
        logger.error('invalid training configuration file: {}'.format(e))
        click.echo('invalid training configuration file.')
        sys.exit(ERROR_EXITCODE)
    except Exception as e:
        logger.error('update job definition version operation aborted: {}'.format(e))
        click.echo('update job definition version operation aborted.')
        sys.exit(ERROR_EXITCODE)

    params = {}
    if description is not None:
        params['description'] = description

    try:
        if version is None:
            try:
                version = _get_latest_training_version(config['name'])
            except ResourceNotFound:
                logger.error('there is no available training versions.')
                click.echo(
                    'there is no available training versions. please create training version first.')
                sys.exit(ERROR_EXITCODE)

        url = "{}/training/definitions/{}/versions/{}".format(
            ORGANIZATION_ENDPOINT, config['name'], version)
        r = api_patch(url, json.dumps(params))
        click.echo(json_output_formatter(r))
    except Exception as e:
        logger.error('update job definition version operation aborted: {}'.format(e))
        click.echo('update job definition version operation aborted.')
        sys.exit(ERROR_EXITCODE)


@training.command(name='describe-versions')
def describe_training_versions():
    try:
        config_data = training_config.read(training_config.default_schema)
        r = _describe_training_versions(config_data['name'])
    except ConfigFileNotFoundError:
        logger.error('training configuration file does not exists.')
        click.echo('training configuration file does not exists.')
        sys.exit(ERROR_EXITCODE)
    except InvalidConfigException as e:
        logger.error('invalid training configuration file: {}'.format(e))
        click.echo('invalid training configuration file.')
        sys.exit(ERROR_EXITCODE)
    except Exception as e:
        logger.error('describe job operation aborted: {}'.format(e))
        click.echo('describe job operation aborted.')
        sys.exit(ERROR_EXITCODE)
    click.echo(json_output_formatter(r))


def _describe_training_versions(name: str):
    url = "{}/training/definitions/{}/versions".format(
        ORGANIZATION_ENDPOINT, name)
    return api_get(url)


def _get_latest_training_version(name: str):
    res = _describe_training_versions(name)
    versions = res.get('entries', [])
    # sort by job_definition_version
    sorted_versions = sorted(versions, key=itemgetter(
        'job_definition_version'), reverse=True)
    if not sorted_versions:
        raise ResourceNotFound
    return sorted_versions[0]['job_definition_version']


@training.command(name='create-job', help='Create training job')
@click.option('-v', '--version', type=str, required=False,
              help='Job definition version. By default, latest version is used')
@click.option('-p', '--params', type=USER_PARAM_STR, default=None, required=False, multiple=True,
              help='User parameters, ex. BATCH_SIZE:32')
@click.option('--instance-type', type=str, required=False,
              help='Instance Type of the machine where training job is executed. '
                   'By default, cpu-1 and gpu-1 is used for all-cpu and all-gpu images respectively.')
@click.option('-d', '--description', type=str, required=False,
              help='Description for the training job, which must be less than or equal to 256 characters.')
def create_training_job(version, params, instance_type, description):
    try:
        config_data = training_config.read(training_config.default_schema)
        if version is None:
            try:
                version = _get_latest_training_version(config_data['name'])
            except ResourceNotFound:
                logger.error('there is no available training versions.')
                click.echo(
                    'there is no available training versions. please create training version first.')
                sys.exit(ERROR_EXITCODE)

        url = "{}/training/definitions/{}/versions/{}/jobs".format(
            ORGANIZATION_ENDPOINT, config_data['name'], version)
        user_params = dict((x, y) for x, y in params)
        data = {'user_parameters': user_params}
        if instance_type is not None:
            data['instance_type'] = instance_type
        if description is not None:
            data['description'] = description
        r = api_post(url, json.dumps(data))
    except ConfigFileNotFoundError:
        logger.error('training configuration file does not exists.')
        click.echo('training configuration file does not exists.')
        sys.exit(ERROR_EXITCODE)
    except InvalidConfigException as e:
        logger.error('invalid training configuration file: {}'.format(e))
        click.echo('invalid training configuration file.')
        sys.exit(ERROR_EXITCODE)
    except Exception as e:
        logger.error('create training job aborted: {}'.format(e))
        click.echo('create training job aborted.')
        sys.exit(ERROR_EXITCODE)
    click.echo(json_output_formatter(r))


@training.command(name='describe-jobs', help='Show training job list for a specific training definition name')
def describe_jobs():
    try:
        config_data = training_config.read(training_config.default_schema)
        url = "{}/training/definitions/{}/jobs".format(
            ORGANIZATION_ENDPOINT, config_data['name'])
        r = api_get(url)
    except ConfigFileNotFoundError:
        logger.error('training configuration file does not exists.')
        click.echo('training configuration file does not exists.')
        sys.exit(ERROR_EXITCODE)
    except InvalidConfigException as e:
        logger.error('invalid training configuration file: {}'.format(e))
        click.echo('invalid training configuration file.')
        sys.exit(ERROR_EXITCODE)
    except Exception as e:
        logger.error('describe job operation aborted: {}'.format(e))
        click.echo('describe job operation aborted.')
        sys.exit(ERROR_EXITCODE)
    click.echo(json_output_formatter(r))


@training.command(name='download-training-result', help='Download training result(artifact)')
@click.option('-jd', '--job_definisin_id', 'job_definisin_id', type=str, help='job definition id', required=True)
@click.option('-j', '--job_id', 'job_id', type=str, help='job id', required=True)
def download_jobs_result(job_definisin_id, job_id):
    try:
        r = _download_result(job_definisin_id, job_id)
    except:
        sys.exit(ERROR_EXITCODE)

    click.echo('Downloaded a file {}.'.format(r))


def _download_result(job_definition_id, job_id):

    url = '{}/training/definitions/{}/jobs/{}/result'.format(
        ORGANIZATION_ENDPOINT, job_definition_id, job_id)
    r = api_get(url)
    download_url = r['artifacts']['complete']['uri']
    file_name = os.path.basename(download_url[:download_url.find('?')])
    target_file_name = '{}_{}_{}'.format(job_definition_id, job_id, file_name)
    urllib.request.urlretrieve(
        download_url, target_file_name, progress_status)

    return target_file_name


def __training_config_provider(yaml_path, _cmd_name):
    try:
        config = read_training_config(yaml_path)
        if not config:
            return {}

        datasets = [
            '{}:{}'.format(k, v)
            for k, v in config.pop('datasets', {}).items()
        ]
        config['datasets'] = datasets

        environment = []
        for k, v in config.pop('params', {}).items():
            if v is None:
                v = ''
            environment.append('{}:{}'.format(k, v))
        config['environment'] = environment
        return config
    except ConfigFileNotFoundError:
        return {}
    except InvalidConfigException as e:
        logger.error('invalid training configuration file: {}'.format(e))
        click.echo('invalid training configuration file.')
        sys.exit(ERROR_EXITCODE)
    except Exception as e:
        logger.error('something wrong occurred when loading training configuration file: {}'.format(e))
        click.echo('something wrong occurred when loading training configuration file.')
        sys.exit(ERROR_EXITCODE)


@training.command(name='archive-job', help='Archive training job')
@click.option('--job-id', '--job_id', type=str, required=True,
              help='Training Job identifier')
def archive_job(job_id):
    try:
        config_data = training_config.read(training_config.default_schema)
        url = "{}/training/definitions/{}/jobs/{}/archive".format(
            ORGANIZATION_ENDPOINT, config_data['name'], job_id)
        r = api_post(url)
    except ConfigFileNotFoundError:
        logger.error('training configuration file does not exists.')
        click.echo('training configuration file does not exists.')
        sys.exit(ERROR_EXITCODE)
    except InvalidConfigException as e:
        logger.error('invalid training configuration file: {}'.format(e))
        click.echo('invalid training configuration file.')
        sys.exit(ERROR_EXITCODE)
    except Exception as e:
        logger.error('create training job aborted: {}'.format(e))
        click.echo('create training job aborted.')
        sys.exit(ERROR_EXITCODE)
    click.echo(json_output_formatter(r))


@training.command(name='unarchive-job', help='Unarchive training job')
@click.option('--job-id', '--job_id', type=str, required=True,
              help='Training Job identifier')
def unarchive_job(job_id):
    try:
        config_data = training_config.read(training_config.default_schema)
        url = "{}/training/definitions/{}/jobs/{}/unarchive".format(
            ORGANIZATION_ENDPOINT, config_data['name'], job_id)
        r = api_post(url)
    except ConfigFileNotFoundError:
        logger.error('training configuration file does not exists.')
        click.echo('training configuration file does not exists.')
        sys.exit(ERROR_EXITCODE)
    except InvalidConfigException as e:
        logger.error('invalid training configuration file: {}'.format(e))
        click.echo('invalid training configuration file.')
        sys.exit(ERROR_EXITCODE)
    except Exception as e:
        logger.error('create training job aborted: {}'.format(e))
        click.echo('create training job aborted.')
        sys.exit(ERROR_EXITCODE)
    click.echo(json_output_formatter(r))


@training.command(name='archive-version', help='Archive training job definition version')
@click.option('-v', '--version-id', '--version_id', type=str, required=True,
              help='Training job version identifier')
def archive_version(version_id):
    try:
        config_data = training_config.read(training_config.default_schema)
        url = "{}/training/definitions/{}/versions/{}/archive".format(
            ORGANIZATION_ENDPOINT, config_data['name'], version_id)
        r = api_post(url)
    except ConfigFileNotFoundError:
        logger.error('training configuration file does not exists.')
        click.echo('training configuration file does not exists.')
        sys.exit(ERROR_EXITCODE)
    except InvalidConfigException as e:
        logger.error('invalid training configuration file: {}'.format(e))
        click.echo('invalid training configuration file.')
        sys.exit(ERROR_EXITCODE)
    except Exception as e:
        logger.error('create training job aborted: {}'.format(e))
        click.echo('create training job aborted.')
        sys.exit(ERROR_EXITCODE)
    click.echo(json_output_formatter(r))


@training.command(name='unarchive-version', help='Unarchive training job definition version')
@click.option('-v', '--version-id', '--version_id', type=str, required=True,
              help='Training job version identifier')
def unarchive_version(version_id):
    try:
        config_data = training_config.read(training_config.default_schema)
        url = "{}/training/definitions/{}/versions/{}/unarchive".format(
            ORGANIZATION_ENDPOINT, config_data['name'], version_id)
        r = api_post(url)
    except ConfigFileNotFoundError:
        logger.error('training configuration file does not exists.')
        click.echo('training configuration file does not exists.')
        sys.exit(ERROR_EXITCODE)
    except InvalidConfigException as e:
        logger.error('invalid training configuration file: {}'.format(e))
        click.echo('invalid training configuration file.')
        sys.exit(ERROR_EXITCODE)
    except Exception as e:
        logger.error('create training job aborted: {}'.format(e))
        click.echo('create training job aborted.')
        sys.exit(ERROR_EXITCODE)
    click.echo(json_output_formatter(r))


def __try_get_organization_id(ctx, param, value):
    if value is None:
        organization_id = get_organization_id()
        if organization_id is None:
            raise MissingParameter(ctx=ctx, param=param)
        return organization_id
    return value


@training.command(name='debug-local', help='Local train commands', context_settings=dict(
    ignore_unknown_options=True, allow_extra_args=True))
@click.option('-h', '--handler', 'handler', type=str, help='Training handler', required=True)
@click.option('-i', '--image', 'image', type=str, required=True,
              help='Specify base image name and tag in the "name:tag" format')
@click.option('-o', '--organization_id', 'organization_id', type=str, required=False,
              help='Organization ID, organization_id of current credential organization is used by default [required]',
              callback=__try_get_organization_id)
@click.option('-d', '--datasets', 'datasets', type=DATASET_PARAM_STR, help='Datasets name', default=None,
              required=False, multiple=True)
@click.option('-e', '--environment', type=ENVIRONMENT_STR, help='Environment variables', default=None,
              required=False, multiple=True)
@click.option('-v', '--volume', type=VOLUME_PARAM_STR,
              default=None, required=False, multiple=True,
              help='Volume driver options, ex) /path/source/on/host:/path/destination/on/container')
@click.option('--no-cache', '--no_cache', is_flag=True, type=bool, required=False,
              help='Not use built cache')
@click.option('--v1', is_flag=True, type=bool, help='Specify if you use old custom runtime image', default=False,
              required=False)
@click.option('--runtime', 'runtime', type=str, required=False,
              help='Runtime, equivalent to docker run `--runtime` option')
@click.option('--build-only', is_flag=True, type=bool,
              help='Build a docker image only. Not run train command.', required=False)
@click.option('-q', '--quiet', is_flag=True, type=bool, help='Suppress info logs', required=False)
@click_config_file.configuration_option(
    provider=__training_config_provider, implicit=True,
    default=CONFIGFILE_NAME,  # by setting `default`, not allow to use `click.get_app_dir`
    config_file_name=CONFIGFILE_NAME,
    help='Read Configuration from PATH. By default read from `{}`'.format(CONFIGFILE_NAME))
def debug_local(
        handler, image, organization_id, datasets, environment, volume,
        no_cache, quiet, v1, runtime=None, build_only=False):
    datasets = dict(datasets) if datasets else {}
    environment = dict(environment) if environment else {}
    volume = build_volumes(volume) if volume else {}
    try:
        with TrainingJobDebugRun(
            handler=handler, image=image, organization_id=organization_id,
            datasets=datasets, environment=environment, volume=volume, no_cache=no_cache,
            build_only=build_only, quiet=quiet, stdout=click.echo, runtime=runtime,
            platform_user_id=ABEJA_PLATFORM_USER_ID,
            platform_personal_access_token=ABEJA_PLATFORM_TOKEN, v1flag=v1
        ) as job:
            if job.build_only:
                sys.exit(SUCCESS_EXITCODE)
            job.watch()
    except Exception as e:
        logger.error('debug local aborted: {}'.format(e))
        click.echo('debug local aborted.')
        sys.exit(ERROR_EXITCODE)


@training.command(name='train-local', help='Local train commands', context_settings=dict(
    ignore_unknown_options=True, allow_extra_args=True))
@click.option('-o', '--organization_id', '--organization-id', 'organization_id', type=str, required=False,
              help='Organization ID, organization_id of current credential organization is used by default [required]',
              callback=__try_get_organization_id)
@click.option('--name', 'name', type=str, help='Training Job Definition Name', required=True)
@click.option('--version', 'version', type=str, help='Training Job Definition Version', required=True)
@click.option('--description', 'description', type=str, help='Training Job description', required=False)
@click.option('-d', '--datasets', 'datasets', type=DATASET_PARAM_STR, help='Datasets name', default=None,
              required=False, multiple=True)
@click.option('-e', '--environment', type=ENVIRONMENT_STR, help='Environment variables', default=None,
              required=False, multiple=True)
@click.option('-v', '--volume', type=VOLUME_PARAM_STR,
              default=None, required=False, multiple=True,
              help='Volume driver options, ex) /path/source/on/host:/path/destination/on/container')
@click.option('--v1', is_flag=True, type=bool, help='Specify if you use old custom runtime image', default=False,
              required=False)
@click.option('--runtime', 'runtime', type=str, required=False,
              help='Runtime, equivalent to docker run `--runtime` option')
@click_config_file.configuration_option(
    provider=__training_config_provider, implicit=True,
    default=CONFIGFILE_NAME,  # by setting `default`, not allow to use `click.get_app_dir`
    config_file_name=CONFIGFILE_NAME,
    help='Read Configuration from PATH. By default read from `{}`'.format(CONFIGFILE_NAME))
def train_local(organization_id, name, version, description, datasets, environment, volume, v1, runtime=None):
    datasets = dict(datasets) if datasets else {}
    environment = dict(environment) if environment else {}
    volume = build_volumes(volume) if volume else {}
    try:
        with TrainingJobLocalContainerRun(
            organization_id=organization_id, job_definition_name=name,
            job_definition_version=version, description=description,
            datasets=datasets, environment=environment, volume=volume, runtime=runtime,
            stdout=click.echo, platform_user_id=ABEJA_PLATFORM_USER_ID,
            platform_personal_access_token=ABEJA_PLATFORM_TOKEN, v1flag=v1
        ) as job:
            job.watch()
    except Exception as e:
        logger.error('train local aborted: {}'.format(e))
        click.echo('train local aborted.')
        sys.exit(ERROR_EXITCODE)
